/* Amazon echo to Sonoff
 *  Original code from: https://github.com/kakopappa to emulate Belkin WeMo with esp8266
 *  Originally used NodeMCU variant of ESP8266 but this is largely irrelevant.
 *  Original code:https://github.com/kakopappa/arduino-esp8266-alexa-wemo-switch
 *  Modified by Ian to add configuration web page - essential with Sonoff
 *  My target is the iTead Sonoff device which provides a complete WiFi switch
 *  Instructions on how to program Sonoff: https://tzapu.com/sonoff-firmware-boilerplate-tutorial/
 *
 *  In operation the device starts up & attempts to connect to a previous known network which is remembered by the ESP8266
 *  If it fails, it starts an Access Point (AP) called 'echoXXXX'
 *  Connect to echoXXXX and navigate to 192.168.4.1
 *  Choose your network & a device name. Submit & it should reboot & connect to your newtwork as a station.
 *  If it stutters - cycle the power & all should be well
 *  The device name is held in EEPROM
 *  31/10/2016
 */

#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <WiFiUdp.h>
#include <EEPROM.h>
#include <ArduinoOTA.h>
#include <FS.h>


void prepareIds();
boolean connectWifi();
boolean connectUDP();
void startHttpServer();
void turnOnRelay();
void turnOffRelay();

String wap = "echo";  //AP name is echoXXXX, where XXXX is unique name (chip_id) of ESP8266
String soft_ap_pwd = "";  //Use at least 8 characters OR leave blank for open access - only used for configuration - blank should be ok
String name = "fake wemo"; //what we will call it - initialise to a default

unsigned int chip_id;          //unique identifier

unsigned int localPort = 1900;      // local port to listen on

WiFiUDP UDP;
boolean udpConnected = false;
IPAddress ipMulti(239, 255, 255, 250);
unsigned int portMulti = 1900;      // local port to listen on

ESP8266WebServer server(80);

boolean wifiConnected = false;

char packetBuffer[UDP_TX_PACKET_MAX_SIZE]; //buffer to hold incoming packet,

String serial;
String persistent_uuid;
String device_name;
//------------------------------------------------------------------------
const int relayPin = 12;    //Sonoff relay is pin 12, Nodemcu led is 16
const int ledPin = 13;  //Sonoff led is 13 - indicates AP/STA modes. Could do more.
//------------------------------------------------------------------------
void setup() {
  SPIFFS.begin();   //start the file system
  pinMode(ledPin, OUTPUT);  //LED is on when in AP mode
  led_off();    //initialise
  WiFi.begin();
  ArduinoOTA.begin();

  EEPROM.begin(100);   //reserve 100 bytes - used for saving device name
  //server.on("/favicon.ico", not_found);   //favicon.ico in spiffs now
  server.on("/net", net);   //network info & config
  server.on("/", net);  //root points to config too
  server.onNotFound(handleOther);   //currently this looks after form input from network setup page & spiffs data
  chip_id = ESP.getChipId();  //uniquely identifies device
  Serial.begin(115200);
  Serial.println("\r\nStarted");
  // Setup Relay
  pinMode(relayPin, OUTPUT);
  turnOffRelay();
  check_name();     //see if a name is already stored in EEPROM & assign to 'name'
  Serial.print("Device name : "); Serial.println(name);
  prepareIds();

  delay(5000);  //chance to remember previous connection, yields to ESP8266 network stuff

  if (station())    //we have successfully connected to a network as a station

  {
    wifiConnected = true;     //flag
    WiFi.mode(WIFI_STA);      //station only
    Serial.print("Already connected to access point ");
    Serial.println(WiFi.SSID());    //currently connected to blah
    Serial.print("IP address : "); Serial.println(WiFi.localIP());
    udpConnected = connectUDP();

    if (udpConnected)
    {
      startHttpServer();
    }
  }
  else
  {
    wap = "echo" + String(chip_id);  //unique identifier - start up access point if unable to join a newtork
    WiFi.softAP(wap.c_str(), soft_ap_pwd.c_str());   //give it a name & password
    Serial.print("WAP name: "); Serial.println(wap);
    Serial.println("Access point started - join & go to 192.168.4.1");
    server.begin();   //start server
    Serial.println("HTTP server started");
    led_on(); //indicate access point available
  }

}
//------------------------------------------------------------------------

void loop() {
  ArduinoOTA.handle();

  server.handleClient();
  delay(1);

  // if there's data available, read a packet
  // check if the WiFi and UDP connections were successful
  if (wifiConnected) {
    if (udpConnected) {
      // if there’s data available, read a packet
      int packetSize = UDP.parsePacket();

      if (packetSize) {
        Serial.println("");
        Serial.print("Received packet of size ");
        Serial.println(packetSize);
        Serial.print("From ");
        IPAddress remote = UDP.remoteIP();

        for (int i = 0; i < 4; i++) {
          Serial.print(remote[i], DEC);
          if (i < 3) {
            Serial.print(".");
          }
        }

        Serial.print(", port ");
        Serial.println(UDP.remotePort());

        int len = UDP.read(packetBuffer, 255);

        if (len > 0) {
          packetBuffer[len] = 0;
        }

        String request = packetBuffer;
        //Serial.println("Request:");
        //Serial.println(request);

        if (request.indexOf('M-SEARCH') > 0) {
          if (request.indexOf("urn:Belkin:device:**") > 0) {
            Serial.println("Responding to search request ...");
            respondToSearch();
          }
        }
      }

      delay(10);
    }
  } else {
    // Turn on/off to indicate cannot connect .. Maybe use LED?
  }
}
// -----------------------------------------------------------------------------
void prepareIds() {
  uint32_t chipId = ESP.getChipId();
  char uuid[64];
  sprintf_P(uuid, PSTR("38323636-4558-4dda-9188-cda0e6%02x%02x%02x"),
            (uint16_t) ((chipId >> 16) & 0xff),
            (uint16_t) ((chipId >>  8) & 0xff),
            (uint16_t)   chipId        & 0xff);

  serial = String(uuid);
  persistent_uuid = "Socket-1_0-" + serial;
  device_name = String(name);    //get this from web page or EEPROM rather than hard coding
}
// -----------------------------------------------------------------------------
void respondToSearch() {
  Serial.println("");
  Serial.print("Sending response to ");
  Serial.println(UDP.remoteIP());
  Serial.print("Port : ");
  Serial.println(UDP.remotePort());

  IPAddress localIP = WiFi.localIP();
  char s[16];
  sprintf(s, "%d.%d.%d.%d", localIP[0], localIP[1], localIP[2], localIP[3]);

  String response =
    "HTTP/1.1 200 OK\r\n"
    "CACHE-CONTROL: max-age=86400\r\n"
    "DATE: Fri, 15 Apr 2016 04:56:29 GMT\r\n" // from whence cometh?
    "EXT:\r\n"
    "LOCATION: http://" + String(s) + ":80/setup.xml\r\n"
    "OPT: \"http://schemas.upnp.org/upnp/1/0/\"; ns=01\r\n"
    "01-NLS: b9200ebb-736d-4b93-bf03-835149d13983\r\n"
    "SERVER: Unspecified, UPnP/1.0, Unspecified\r\n"
    "ST: urn:Belkin:device:**\r\n"
    "USN: uuid:" + persistent_uuid + "::urn:Belkin:device:**\r\n"
    "X-User-Agent: redsonic\r\n\r\n";

  UDP.beginPacket(UDP.remoteIP(), UDP.remotePort());
  UDP.write(response.c_str());
  UDP.endPacket();

  Serial.println("Response sent !");
}
// -----------------------------------------------------------------------------
void startHttpServer() {    //this is odd! Replace this with /net response?
  // server.on("/", HTTP_GET, []() {
  //   Serial.println("Got Request index.html ...\n");
  //   server.send(200, "text/plain", "Hello World!");
  //  });

  server.on("/upnp/control/basicevent1", HTTP_POST, []() {
    Serial.println("########## Responding to  /upnp/control/basicevent1 ... ##########");

    String request = server.arg(0);
    Serial.print("request:");
    Serial.println(request);

    if (request.indexOf("<BinaryState>1</BinaryState>") > 0) {
      Serial.println("Got Turn on request");
      turnOnRelay();
    }

    if (request.indexOf("<BinaryState>0</BinaryState>") > 0) {
      Serial.println("Got Turn off request");
      turnOffRelay();
    }

    server.send(200, "text/plain", "");
  });

  server.on("/eventservice.xml", HTTP_GET, []() {
    Serial.println(" ########## Responding to eventservice.xml ... ########\n");
    String eventservice_xml = "<?scpd xmlns=\"urn:Belkin:service-1-0\"?>"
                              "<actionList>"
                              "<action>"
                              "<name>SetBinaryState</name>"
                              "<argumentList>"
                              "<argument>"
                              "<retval/>"
                              "<name>BinaryState</name>"
                              "<relatedStateVariable>BinaryState</relatedStateVariable>"
                              "<direction>in</direction>"
                              "</argument>"
                              "</argumentList>"
                              "<serviceStateTable>"
                              "<stateVariable sendEvents=\"yes\">"
                              "<name>BinaryState</name>"
                              "<dataType>Boolean</dataType>"
                              "<defaultValue>0</defaultValue>"
                              "</stateVariable>"
                              "<stateVariable sendEvents=\"yes\">"
                              "<name>level</name>"
                              "<dataType>string</dataType>"
                              "<defaultValue>0</defaultValue>"
                              "</stateVariable>"
                              "</serviceStateTable>"
                              "</action>"
                              "</scpd>\r\n"
                              "\r\n";

    server.send(200, "text/plain", eventservice_xml.c_str());
  });

  server.on("/setup.xml", HTTP_GET, []() {
    Serial.println(" ########## Responding to setup.xml ... ########\n");

    IPAddress localIP = WiFi.localIP();
    char s[16];
    sprintf(s, "%d.%d.%d.%d", localIP[0], localIP[1], localIP[2], localIP[3]);

    String setup_xml = "<?xml version=\"1.0\"?>"
                       "<root>"
                       "<device>"
                       "<deviceType>urn:Belkin:device:controllee:1</deviceType>"
                       "<friendlyName>" + device_name + "</friendlyName>"
                       "<manufacturer>Belkin International Inc.</manufacturer>"
                       "<modelName>Emulated Socket</modelName>"
                       "<modelNumber>3.1415</modelNumber>"
                       "<UDN>uuid:" + persistent_uuid + "</UDN>"
                       "<serialNumber>221517K0101769</serialNumber>"
                       "<binaryState>0</binaryState>"
                       "<serviceList>"
                       "<service>"
                       "<serviceType>urn:Belkin:service:basicevent:1</serviceType>"
                       "<serviceId>urn:Belkin:serviceId:basicevent1</serviceId>"
                       "<controlURL>/upnp/control/basicevent1</controlURL>"
                       "<eventSubURL>/upnp/event/basicevent1</eventSubURL>"
                       "<SCPDURL>/eventservice.xml</SCPDURL>"
                       "</service>"
                       "</serviceList>"
                       "</device>"
                       "</root>\r\n"
                       "\r\n";

    server.send(200, "text/xml", setup_xml.c_str());

    Serial.print("Sending :");
    Serial.println(setup_xml);
  });

  server.begin();
  Serial.println("HTTP Server started ..");
}

// -----------------------------------------------------------------------------
boolean connectUDP() {
  boolean state = false;

  Serial.println("");
  Serial.println("Connecting to UDP");

  if (UDP.beginMulticast(WiFi.localIP(), ipMulti, portMulti)) {
    Serial.println("Connection successful");
    state = true;
  }
  else {
    Serial.println("Connection failed");
  }
  return state;
}
// -----------------------------------------------------------------------------
// active high or active low - that is the question...
void turnOnRelay() {
  digitalWrite(relayPin, HIGH); // turn on relay with voltage HIGH (LED is low on Node_mcu)
}
// -----------------------------------------------------------------------------
void turnOffRelay() {
  digitalWrite(relayPin, LOW);  // turn off relay with voltage LOW
}

void led_on() {
  digitalWrite(ledPin, 0);
}
void led_off() {
  digitalWrite(ledPin, 1);
}

